/*
*
* SPDX-License-Identifier: BSD-3-Clause
*
* Copyright (c) 2016-2022, Arm Limited. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification,
* are permitted provided that the following conditions are met:
* - Redistributions of source code must retain the above copyright notice, this
*   list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice, this
*    list of conditions and the following disclaimer in the documentation and/or
*    other materials provided with the distribution.
* - Neither the name of ARM nor the names of its contributors may be used to
*    endorse or promote products derived from this software without specific prior
*    written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
* ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
*/

#include "acamera_fw.h"
#include "ae_acamera_fsm.h"

#if defined( ISP_HAS_USER2KERNEL_FSM )
#include "sbuf.h"
#endif

#ifdef LOG_MODULE
#undef LOG_MODULE
#define LOG_MODULE LOG_MODULE_AE_ACAMERA
#endif

void ae_acamera_fsm_clear( ae_acamera_fsm_t *p_fsm )
{
    p_fsm->exposure_log2 = 0;
    p_fsm->exposure_ratio = 64;
    p_fsm->ae_roi_api = AE_CENTER_ZONES;
    p_fsm->roi = AE_CENTER_ZONES;

#if FW_ALG_SUPPORT_SHARED_LIB
    p_fsm->lib_handle = NULL;
#endif
    memset( &p_fsm->ae_alg_obj, 0, sizeof( p_fsm->ae_alg_obj ) );
}

void ae_acamera_request_interrupt( ae_acamera_fsm_ptr_t p_fsm, system_fw_interrupt_mask_t mask )
{
    acamera_isp_interrupts_disable( ACAMERA_FSM2MGR_PTR( p_fsm ) );
    p_fsm->mask.irq_mask |= mask;
    acamera_isp_interrupts_enable( ACAMERA_FSM2MGR_PTR( p_fsm ) );
}

void ae_acamera_fsm_init( void *fsm, fsm_init_param_t *init_param )
{
    ae_acamera_fsm_t *p_fsm = (ae_acamera_fsm_t *)fsm;
    p_fsm->cmn.p_fsm_mgr = init_param->p_fsm_mgr;
    p_fsm->cmn.isp_base = init_param->isp_base;
    p_fsm->p_fsm_mgr = init_param->p_fsm_mgr;

    ae_acamera_fsm_clear( p_fsm );

    ae_acamera_initialize( p_fsm );

    ae_acamera_request_interrupt( p_fsm, ACAMERA_IRQ_MASK( ACAMERA_IRQ_AE_STATS ) );
}

void ae_acamera_fsm_deinit( void *fsm )
{
    ae_acamera_fsm_t *p_fsm = (ae_acamera_fsm_t *)fsm;
    ae_acamera_deinitialize( p_fsm );
}


int ae_acamera_fsm_set_param( void *fsm, uint32_t param_id, void *input, uint32_t input_size )
{
    int rc = 0;
    ae_acamera_fsm_t *p_fsm = (ae_acamera_fsm_t *)fsm;

    switch ( param_id ) {
    case FSM_PARAM_SET_AE_INIT:
        // ae_acamera_fsm_clear( p_fsm );
        // ae_acamera_initialize( p_fsm );
        ae_roi_update( p_fsm );
        break;

#if defined( ISP_HAS_USER2KERNEL_FSM )
    case FSM_PARAM_SET_AE_STATS: {
        if ( !input || input_size != sizeof( sbuf_ae_t ) ) {
            LOG( LOG_ERR, "Invalid param, param_id: %d.", param_id );
            rc = -1;
            break;
        }

        sbuf_ae_t *p_sbuf_ae = (sbuf_ae_t *)input;

        uint32_t len = sizeof( p_sbuf_ae->stats_data );
#if DEBUG_SBUF_AE
        {
            extern int sched_getcpu( void );
            static uint32_t pre_id = 0xFFFFFFFF;
            int i;
            uint32_t histsum = 0;
            LOG( LOG_DEBUG, "len: %u, p_sbuf_ae: %p.", len, p_sbuf_ae );
            for ( i = 0; i < len / 4; i++ ) {
                histsum += p_sbuf_ae->stats_data[i];
                if ( !( i & 0x7F ) ) {
                    LOG( LOG_DEBUG, "fullhist[%d]: %u.", i, p_sbuf_ae->stats_data[i] );
                }
            }

            if ( histsum != p_sbuf_ae->histogram_sum ) {
                LOG( LOG_ERR, "Error: report histgram(%u) is not match with data calculated(%u).", p_sbuf_ae->histogram_sum, histsum );
            }

            if ( pre_id != 0xFFFFFFFF && ( p_sbuf_ae->frame_id - pre_id ) != 1 ) {
                LOG( LOG_ERR, "cpu: %d, pre_id(%u) is not match current frame_id(%u).", sched_getcpu(), pre_id, p_sbuf_ae->frame_id );
            }

            pre_id = p_sbuf_ae->frame_id;
        }
#endif
        system_memcpy( p_fsm->fullhist, p_sbuf_ae->stats_data, len );
        system_memcpy( p_fsm->hist4, p_sbuf_ae->hist4, sizeof( p_fsm->hist4 ) );
        /* The last element is the histogram sum */
        p_fsm->fullhist_sum = p_sbuf_ae->histogram_sum;
        fsm_raise_event( p_fsm, event_id_ae_stats_ready );

        break;
    }
#endif

    case FSM_PARAM_SET_AE_ROI: {
        if ( !input || input_size != sizeof( fsm_param_roi_t ) ) {
            LOG( LOG_ERR, "Invalid param, param_id: %d.", param_id );
            rc = -1;
            break;
        }

        fsm_param_roi_t *p_new = (fsm_param_roi_t *)input;
        p_fsm->ae_roi_api = p_new->roi_api;
        p_fsm->roi = p_new->roi;

        ae_roi_update( p_fsm );

        break;
    }

    default:
        rc = -1;
        break;
    }

    return rc;
}


int ae_acamera_fsm_get_param( void *fsm, uint32_t param_id, void *input, uint32_t input_size, void *output, uint32_t output_size )
{
    int rc = 0;
    ae_acamera_fsm_t *p_fsm = (ae_acamera_fsm_t *)fsm;

    switch ( param_id ) {
    case FSM_PARAM_GET_AE_INFO: {
        if ( !output || output_size != sizeof( fsm_param_ae_info_t ) ) {
            LOG( LOG_ERR, "Invalid param, param_id: %d.", param_id );
            rc = -1;
            break;
        }

        fsm_param_ae_info_t *p_ae_info = (fsm_param_ae_info_t *)output;

        p_ae_info->exposure_log2 = p_fsm->exposure_log2;
        p_ae_info->exposure_ratio = p_fsm->exposure_ratio;
        break;
    }

    case FSM_PARAM_GET_AE_HIST_INFO: {
        if ( !output || output_size != sizeof( fsm_param_ae_hist_info_t ) ) {
            LOG( LOG_ERR, "Invalid param, param_id: %d.", param_id );
            rc = -1;
            break;
        }

        fsm_param_ae_hist_info_t *p_hist_info = (fsm_param_ae_hist_info_t *)output;

        p_hist_info->fullhist_sum = p_fsm->fullhist_sum;
        p_hist_info->fullhist = p_fsm->fullhist;
        p_hist_info->fullhist_size = array_size( p_fsm->fullhist );
        p_hist_info->frame_id = p_fsm->cur_using_stats_frame_id;

        break;
    }

    case FSM_PARAM_GET_AE_ROI: {
        if ( !output || output_size != sizeof( fsm_param_roi_t ) ) {
            LOG( LOG_ERR, "Invalid param, param_id: %d.", param_id );
            rc = -1;
            break;
        }

        fsm_param_roi_t *p_current = (fsm_param_roi_t *)output;
        p_current->roi_api = p_fsm->ae_roi_api;
        p_current->roi = p_fsm->roi;

        break;
    }

    case FSM_PARAM_GET_AE_STATE: {
        ae_state_t state;

        if ( !output || output_size != sizeof( ae_state_t ) ) {
            LOG( LOG_ERR, "Invalid param, param_id: %d.", param_id );
            rc = -1;
            break;
        }

        if ( ACAMERA_FSM2CTX_PTR( p_fsm )->stab.global_manual_exposure ) {
            state = AE_STATE_INACTIVE;
        } else {
            if ( p_fsm->ae_converged )
                state = AE_STATE_CONVERGED;
            else
                state = AE_STATE_SEARCHING;
        }

        *(ae_state_t *)output = state;

        break;
    }

    default:
        rc = -1;
        break;
    }

    return rc;
}


uint8_t ae_acamera_fsm_process_event( ae_acamera_fsm_t *p_fsm, event_id_t event_id )
{
    uint8_t b_event_processed = 0;
    switch ( event_id ) {
    default:
        break;
    case event_id_ae_stats_ready:
        ae_process_stats( p_fsm );
        fsm_raise_event( p_fsm, event_id_exposure_changed );
        ae_acamera_request_interrupt( p_fsm, ACAMERA_IRQ_MASK( ACAMERA_IRQ_AE_STATS ) );
        b_event_processed = 1;
        break;
    }

    return b_event_processed;
}
